import multiprocessing
import os
import shutil
import subprocess
import unittest
from glob import glob
from unittest.mock import Mock, patch

import pymatgen
import pytest
from monty.os import cd
from monty.tempfile import ScratchDir
from pymatgen.core import Structure
from pymatgen.io.vasp import Incar, Kpoints, Poscar
from pymatgen.io.vasp.sets import MPRelaxSet

from custodian.vasp.jobs import GenerateVaspInputJob, VaspJob, VaspNEBJob, _gamma_point_only_check
from tests.conftest import TEST_FILES

pymatgen.core.SETTINGS["PMG_VASP_PSP_DIR"] = TEST_FILES


class TestVaspJob:
    def test_as_from_dict(self) -> None:
        v = VaspJob(["hello"])
        v2 = VaspJob.from_dict(v.as_dict())
        assert isinstance(v2, VaspJob)
        assert v2.vasp_cmd == ("hello",)

    def test_setup(self) -> None:
        with cd(TEST_FILES), ScratchDir(".", copy_from_current_on_enter=True):
            v = VaspJob(["hello"], auto_npar=True)
            v.setup()
            incar = Incar.from_file("INCAR")
            count = multiprocessing.cpu_count()
            # Need at least 3 CPUs for NPAR to be greater than 1
            if count > 3:
                assert incar["NPAR"] > 1

    def test_setup_run_no_kpts(self) -> None:
        # just make sure v.setup() and v.run() exit cleanly when no KPOINTS file is present
        with cd(f"{TEST_FILES}/kspacing"), ScratchDir(".", copy_from_current_on_enter=True):
            v = VaspJob(["hello"], auto_npar=True)
            v.setup()
            with pytest.raises(FileNotFoundError):
                # a FileNotFoundError indicates that v.run() tried to run
                # subprocess.Popen(cmd, stdout=f_std, stderr=f_err) with
                # cmd == "hello", so it successfully parsed the input file
                # directory.
                v.run()

    def test_update_incar(self) -> None:
        # just make sure v.setup() and v.run() exit cleanly when no KPOINTS file is present
        with cd(f"{TEST_FILES}/update_incar"), ScratchDir(".", copy_from_current_on_enter=True):
            incar = Incar.from_file("INCAR")
            assert incar["NBANDS"] == 200
            v = VaspJob(["hello"], update_incar=True)
            v.setup()
            incar = Incar.from_file("INCAR")
            assert incar["NBANDS"] == 224

    def test_postprocess(self) -> None:
        with cd(f"{TEST_FILES}/postprocess"), ScratchDir(".", copy_from_current_on_enter=True):
            shutil.copy("INCAR", "INCAR.backup")

            v = VaspJob(["hello"], final=False, suffix=".test", copy_magmom=True)
            v.postprocess()
            incar = Incar.from_file("INCAR")
            incar_prev = Incar.from_file("INCAR.test")

            for file in (
                "INCAR",
                "KPOINTS",
                "CONTCAR",
                "OSZICAR",
                "OUTCAR",
                "POSCAR",
                "vasprun.xml",
            ):
                assert os.path.isfile(f"{file}.test")
                os.remove(f"{file}.test")
            shutil.move("INCAR.backup", "INCAR")

            assert incar["MAGMOM"] == pytest.approx([3.007, 1.397, -0.189, -0.189])
            assert incar_prev["MAGMOM"] == pytest.approx([5, -5, 0.6, 0.6])

    def test_continue(self) -> None:
        # Test the continuation functionality
        with cd(f"{TEST_FILES}/postprocess"):
            # Test default functionality
            with ScratchDir(".", copy_from_current_on_enter=True):
                v = VaspJob("hello", auto_continue=True)
                v.setup()
                assert os.path.isfile("continue.json"), "continue.json not created"
                v.setup()
                assert Poscar.from_file("CONTCAR").structure == Poscar.from_file("POSCAR").structure
                assert Incar.from_file("INCAR")["ISTART"] == 1
                v.postprocess()
                assert not os.path.isfile("continue.json"), "continue.json not deleted after postprocessing"
            # Test explicit action functionality
            with ScratchDir(".", copy_from_current_on_enter=True):
                v = VaspJob(
                    ["hello"],
                    auto_continue=[{"dict": "INCAR", "action": {"_set": {"ISTART": 1}}}],
                )
                v.setup()
                v.setup()
                assert Poscar.from_file("CONTCAR").structure != Poscar.from_file("POSCAR").structure
                assert Incar.from_file("INCAR")["ISTART"] == 1
                v.postprocess()

    def test_static(self) -> None:
        # Just a basic test of init.
        VaspJob.double_relaxation_run(["vasp"])


class TestVaspNEBJob:
    def test_as_from_dict(self) -> None:
        v = VaspNEBJob(["hello"])
        v2 = VaspNEBJob.from_dict(v.as_dict())
        assert isinstance(v2, VaspNEBJob)
        assert v2.vasp_cmd == ("hello",)

    def test_setup(self) -> None:
        with cd(f"{TEST_FILES}/setup_neb"), ScratchDir(".", copy_from_current_on_enter=True):
            v = VaspNEBJob("hello", half_kpts=True)
            v.setup()

            incar = Incar.from_file("INCAR")
            count = multiprocessing.cpu_count()
            if count > 3:
                assert incar["NPAR"] > 1

            kpt = Kpoints.from_file("KPOINTS")
            kpt_pre = Kpoints.from_file("KPOINTS.orig")
            assert kpt_pre.style.name == "Monkhorst"
            assert kpt.style.name == "Gamma"

    def test_postprocess(self) -> None:
        neb_outputs = ["INCAR", "KPOINTS", "POTCAR", "vasprun.xml"]
        neb_sub_outputs = [
            "CHG",
            "CHGCAR",
            "CONTCAR",
            "DOSCAR",
            "EIGENVAL",
            "IBZKPT",
            "PCDAT",
            "POSCAR",
            "REPORT",
            "PROCAR",
            "OSZICAR",
            "OUTCAR",
            "WAVECAR",
            "XDATCAR",
        ]

        with cd(f"{TEST_FILES}/postprocess_neb"):
            postprocess_neb = os.path.abspath(".")

            v = VaspNEBJob("hello", final=False, suffix=".test")
            v.postprocess()

            for file in neb_outputs:
                assert os.path.isfile(f"{file}.test")
                os.remove(f"{file}.test")

            sub_folders = glob("[0-9][0-9]")
            for sf in sub_folders:
                os.chdir(f"{postprocess_neb}/{sf}")
                for file in neb_sub_outputs:
                    if os.path.isfile(file):
                        assert os.path.isfile(f"{file}.test")
                        os.remove(f"{file}.test")


class TestGenerateVaspInputJob:
    def test_run(self) -> None:
        with ScratchDir("."):
            for file in ("INCAR", "POSCAR", "POTCAR", "KPOINTS"):
                shutil.copy(f"{TEST_FILES}/{file}", file)
            old_incar = Incar.from_file("INCAR")
            v = GenerateVaspInputJob("pymatgen.io.vasp.sets.MPNonSCFSet", contcar_only=False)
            v.run()
            incar = Incar.from_file("INCAR")
            assert incar["ICHARG"] == 11
            assert old_incar["ICHARG"] == 1
            kpoints = Kpoints.from_file("KPOINTS")
            assert str(kpoints.style) == "Reciprocal"


class TestAutoGamma:
    """
    Test that a VASP job can automatically detect when only 1 k-point at GAMMA is used.
    """

    def test_gamma_checks(self) -> None:
        # Isolated atom in PBC
        structure = Structure(
            lattice=[[15 + 0.1 * i if i == j else 0.0 for j in range(3)] for i in range(3)],
            species=["Na"],
            coords=[[0.5 for _ in range(3)]],
        )

        vis = MPRelaxSet(structure=structure)
        assert vis.kpoints.kpts == [(1, 1, 1)]
        assert _gamma_point_only_check(vis.get_input_set())

        # no longer Gamma-centered
        vis = MPRelaxSet(structure=structure, user_kpoints_settings=Kpoints(kpts=[2, 1, 1]))
        assert not _gamma_point_only_check(vis.get_input_set())

        vis = MPRelaxSet(structure=structure, user_kpoints_settings=Kpoints(kpts_shift=(0.1, 0.0, 0.0)))
        assert not _gamma_point_only_check(vis.get_input_set())

        # KSPACING-related checks
        vis = MPRelaxSet(structure=structure, user_incar_settings={"KSPACING": 0.005})
        assert not _gamma_point_only_check(vis.get_input_set())

        vis = MPRelaxSet(structure=structure, user_incar_settings={"KSPACING": 50})
        assert _gamma_point_only_check(vis.get_input_set())


class TestVaspJobTerminate(unittest.TestCase):
    """Unit tests for the VaspJob.terminate() method."""

    def setUp(self):
        """Set up test fixtures before each test method."""
        # Create a VaspJob instance with minimal required parameters
        self.vasp_job = VaspJob(vasp_cmd=["srun", "vasp"])

        # Mock the _vasp_process attribute
        self.mock_process = Mock()
        self.vasp_job._vasp_process = self.mock_process

        # Set up logging capture
        self.logger_patcher = patch("custodian.vasp.jobs.logger")
        self.mock_logger = self.logger_patcher.start()

    def tearDown(self):
        """Clean up after each test method."""
        self.logger_patcher.stop()

    def test_terminate_process_already_finished(self):
        """Test termination when process has already finished (poll returns non-None)."""
        # Arrange
        self.mock_process.poll.return_value = 0  # Process already finished

        # Act
        self.vasp_job.terminate()

        # Assert
        self.mock_process.poll.assert_called_once()
        self.mock_logger.warning.assert_called_once_with("The process was already done!")
        self.mock_process.terminate.assert_not_called()
        self.mock_process.kill.assert_not_called()

    def test_terminate_graceful_success(self):
        """Test successful graceful termination."""
        # Arrange
        self.mock_process.poll.return_value = None  # Process is running
        self.mock_process.pid = 12345
        self.mock_process.terminate.return_value = None
        self.mock_process.wait.return_value = None

        # Act
        self.vasp_job.terminate()

        # Assert
        self.mock_process.poll.assert_called_once()
        self.mock_process.terminate.assert_called_once()
        self.mock_process.wait.assert_called_once_with(timeout=10)
        self.mock_logger.info.assert_called_once_with("Killing PID 12345")
        self.mock_process.kill.assert_not_called()

    def test_terminate_graceful_timeout_then_force_kill(self):
        """Test graceful termination timeout leading to force kill."""
        # Arrange
        self.mock_process.poll.return_value = None  # Process is running
        self.mock_process.pid = 12345
        self.mock_process.terminate.return_value = None
        self.mock_process.wait.side_effect = [
            subprocess.TimeoutExpired(cmd="vasp", timeout=10),  # First call times out
            None,  # Second call succeeds
        ]

        # Act
        self.vasp_job.terminate()

        # Assert
        self.mock_process.poll.assert_called_once()
        self.mock_process.terminate.assert_called_once()
        assert self.mock_process.wait.call_count == 2
        self.mock_process.kill.assert_called_once()

        # Check logging calls
        self.mock_logger.info.assert_called_once_with("Killing PID 12345")
        self.mock_logger.warning.assert_called_once_with("Graceful termination did not work. Force killing PID 12345")

    def test_terminate_exception_during_graceful_termination(self):
        """Test handling of exceptions during graceful termination."""
        # Arrange
        self.mock_process.poll.return_value = None
        self.mock_process.pid = 12345
        self.mock_process.terminate.side_effect = OSError("Permission denied")

        # Act & Assert
        with pytest.raises(OSError):
            self.vasp_job.terminate()

        self.mock_process.terminate.assert_called_once()

    def test_terminate_exception_during_force_kill(self):
        """Test handling of exceptions during force kill."""
        # Arrange
        self.mock_process.poll.return_value = None
        self.mock_process.pid = 12345
        self.mock_process.terminate.return_value = None
        self.mock_process.wait.side_effect = [
            subprocess.TimeoutExpired(cmd="vasp", timeout=10),  # Graceful timeout
            OSError("Process not found"),  # Force kill fails
        ]
        self.mock_process.kill.return_value = None

        # Act & Assert
        with pytest.raises(OSError):
            self.vasp_job.terminate()

        self.mock_process.terminate.assert_called_once()
        self.mock_process.kill.assert_called_once()

    def test_terminate_multiple_calls(self):
        """Test calling terminate multiple times."""
        # Arrange
        self.mock_process.poll.side_effect = [None, 0]  # Running, then finished
        self.mock_process.pid = 12345
        self.mock_process.terminate.return_value = None
        self.mock_process.wait.return_value = None

        # Act
        self.vasp_job.terminate()  # First call
        self.vasp_job.terminate()  # Second call

        # Assert
        assert self.mock_process.poll.call_count == 2
        self.mock_process.terminate.assert_called_once()  # Only called on first invocation
        self.mock_logger.warning.assert_called_once_with("The process was already done!")

    @patch("custodian.vasp.jobs.VaspJob.__init__", return_value=None)
    def test_terminate_integration_with_real_process(self, mock_init):
        """Test termination with a real subprocess (integration-style test)."""
        # Arrange
        vasp_job = VaspJob.__new__(VaspJob)  # Create instance without calling __init__

        real_process = subprocess.Popen(["sleep", "10"])
        vasp_job._vasp_process = real_process

        with patch("custodian.vasp.jobs.logger") as mock_logger:
            vasp_job.terminate()
        assert real_process.poll() is not None  # Process should be terminated
        mock_logger.info.assert_called_once()


class TestVaspJobTerminateEdgeCases(unittest.TestCase):
    """Additional edge case tests for VaspJob.terminate()."""

    def setUp(self):
        self.vasp_job = VaspJob(vasp_cmd=["vasp"])
        self.mock_process = Mock()
        self.vasp_job._vasp_process = self.mock_process

    @patch("custodian.vasp.jobs")
    def test_terminate_wait_with_different_timeout_behavior(self, mock_logger):
        """Test different timeout behaviors in wait()."""
        # Test case where wait() is called twice with different outcomes
        self.mock_process.poll.return_value = None
        self.mock_process.pid = 9999
        self.mock_process.terminate.return_value = None

        # First wait times out, second wait after kill succeeds
        self.mock_process.wait.side_effect = [subprocess.TimeoutExpired("vasp", 10), None]

        self.vasp_job.terminate()

        assert self.mock_process.wait.call_count == 2
        self.mock_process.kill.assert_called_once()



# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

import os
import unittest
import warnings
from glob import glob
from pathlib import Path

from custodian.cp2k.jobs import Cp2kJob
from custodian.custodian import Custodian
from tests.conftest import TEST_FILES

MODULE_DIR = Path(__file__).resolve().parent
TEST_FILES_DIR = f"{TEST_FILES}/cp2k"

cwd = os.getcwd()


def clean_dir(folder) -> None:
    for file in glob(os.path.join(folder, "error.*.tar.gz")):
        os.remove(file)
    for file in glob(os.path.join(folder, "custodian.chk.*.tar.gz")):
        os.remove(file)


class HandlerTests(unittest.TestCase):
    def setUp(self) -> None:
        warnings.filterwarnings("ignore")

        clean_dir(TEST_FILES_DIR)

        self.input_file = f"{TEST_FILES_DIR}/cp2k.inp"
        self.input_file_hybrid = f"{TEST_FILES_DIR}/cp2k.inp.hybrid"
        self.output_file = f"{TEST_FILES_DIR}/cp2k.out.test"
        self.std_err = f"{TEST_FILES_DIR}/std_err.tmp"
        self.logfile = f"{TEST_FILES_DIR}/custodian.json"

        if os.path.isfile(Custodian.LOG_FILE):
            os.remove("custodian.json")
        if os.path.isfile(self.std_err):
            os.remove(self.std_err)
        if os.path.isfile(self.output_file):
            os.remove(self.output_file)

    def test_job(self) -> None:
        job = Cp2kJob(
            cp2k_cmd=["echo"],
            input_file=self.input_file,
            output_file=self.output_file,
            stderr_file=self.std_err,
            suffix="",
            final=True,
            backup=False,
            settings_override=None,
        )
        c = Custodian(jobs=[job], handlers=[])
        c.run()
        if os.path.isfile(Custodian.LOG_FILE):
            os.remove("custodian.json")
        if os.path.isfile(self.std_err):
            os.remove(self.std_err)

    def test_double(self) -> None:
        jobs = Cp2kJob.double_job(
            cp2k_cmd=["echo"],
            input_file=self.input_file_hybrid,
            output_file=self.output_file,
            stderr_file="std_err.tmp",
            backup=False,
        )
        assert len(jobs) == 2



"""This module implements specific error handler for FEFF runs."""

from __future__ import annotations

import logging
import os
import re

from pymatgen.io.feff.sets import FEFFDictSet

from custodian.custodian import ErrorHandler
from custodian.feff.interpreter import FeffModder
from custodian.utils import backup

__author__ = "Chen Zheng"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Chen Zheng"
__email__ = "chz022@ucsd.edu"
__date__ = "Oct 18, 2017"

FEFF_BACKUP_FILES = [
    "ATOMS",
    "HEADER",
    "PARAMETERS",
    "POTENTIALS",
    "feff.inp",
    "*.cif",
    "pot.bin",
]

logger = logging.getLogger(__name__)


class UnconvergedErrorHandler(ErrorHandler):
    """Correct the unconverged error of FEFF's SCF calculation."""

    is_monitor = False

    def __init__(self, output_filename="log1.dat") -> None:
        """Initialize the handler with the output file to check.

        Args:
            output_filename (str): Filename for the log1.dat file. log1.dat file
                contains the SCF calculation convergence information. Change this only
                if it is different from the default (unlikely).
        """
        self.output_filename = output_filename

    def check(self, directory="./"):
        """
        If the FEFF run does not converge, the check will return
        "TRUE".
        """
        return self._notconverge_check(directory)

    def _notconverge_check(self, directory) -> bool | None:
        # Process the output file and get converge information
        not_converge_pattern = re.compile("Convergence not reached.*")
        converge_pattern = re.compile("Convergence reached.*")
        with open(os.path.join(directory, self.output_filename)) as file:
            for line in file:
                if len(not_converge_pattern.findall(line)) > 0:
                    return True

                if len(converge_pattern.findall(line)) > 0:
                    return False
        return None

    def correct(self, directory="./"):
        """Perform the corrections."""
        backup(FEFF_BACKUP_FILES, directory=directory)
        feff_input = FEFFDictSet.from_directory(directory)
        scf_values = feff_input.tags.get("SCF")
        nscmt = scf_values[2]
        ca = scf_values[3]
        nmix = scf_values[4]
        actions = []

        # Add RESTART card to PARAMETERS
        if "RESTART" not in feff_input.tags:
            actions += [{"dict": "PARAMETERS", "action": {"_set": {"RESTART": []}}}]

        if nscmt < 100 and ca == 0.2:
            scf_values[2] = 100
            scf_values[4] = 3  # Set nmix = 3
            actions += [{"dict": "PARAMETERS", "action": {"_set": {"SCF": scf_values}}}]
            FeffModder(directory=directory).apply_actions(actions)
            return {"errors": ["Non-converging job"], "actions": actions}

        if nscmt == 100 and nmix == 3 and ca > 0.01:
            # Reduce the convergence accelerator factor
            scf_values[3] = round(ca / 2, 2)
            actions += [{"dict": "PARAMETERS", "action": {"_set": {"SCF": scf_values}}}]
            FeffModder(directory=directory).apply_actions(actions)
            return {"errors": ["Non-converging job"], "actions": actions}

        if nmix == 3 and ca == 0.01:
            # Set ca = 0.05 and set nmix
            scf_values[3] = 0.05
            scf_values[4] = 5
            actions += [{"dict": "PARAMETERS", "action": {"_set": {"SCF": scf_values}}}]
            FeffModder(directory=directory).apply_actions(actions)
            return {"errors": ["Non-converging job"], "actions": actions}

        if nmix == 5 and ca == 0.05:
            # Set ca = 0.05 and set nmix
            scf_values[3] = 0.05
            scf_values[4] = 10
            actions += [{"dict": "PARAMETERS", "action": {"_set": {"SCF": scf_values}}}]
            FeffModder(directory=directory).apply_actions(actions)
            return {"errors": ["Non-converging job"], "actions": actions}

        if nmix == 10 and ca < 0.2:
            # loop through ca with nmix = 10
            scf_values[3] = round(ca * 2, 2)
            actions += [{"dict": "PARAMETERS", "action": {"_set": {"SCF": scf_values}}}]
            FeffModder(directory=directory).apply_actions(actions)
            return {"errors": ["Non-converging job"], "actions": actions}

        # Unfixable error. Just return None for actions.
        return {"errors": ["Non-converging job"], "actions": None}



"""Deployment file to facilitate releases of custodian."""

import datetime
import json
import os
import re
import subprocess
from glob import glob

import requests
from invoke import task
from monty.os import cd

from custodian import __version__ as CURRENT_VER

NEW_VER = datetime.datetime.now().strftime("%Y.%-m.%-d")


@task
def make_doc(ctx) -> None:
    with cd("docs"):
        ctx.run("touch index.rst")
        ctx.run("rm custodian.*.rst", warn=True)
        ctx.run("sphinx-apidoc --separate -P -M -d 7 -o . -f ../src/custodian")
        ctx.run("sphinx-build -M markdown . .")
        ctx.run("rm *.rst", warn=True)
        ctx.run("cp markdown/custodian*.md .")
        ctx.run("rm custodian*tests*.md", warn=True)
        for fn in glob("custodian*.md"):
            with open(fn) as file:
                lines = [line.rstrip() for line in file if "Submodules" not in line]
            if fn == "custodian.md":
                preamble = ["---", "layout: default", "title: API Documentation", "nav_order: 6", "---", ""]
            else:
                preamble = ["---", "layout: default", "title: " + fn, "nav_exclude: true", "---", ""]
            with open(fn, "w") as file:
                file.write("\n".join(preamble + lines))
        ctx.run("rm -r markdown doctrees", warn=True)


@task
def update_doc(ctx) -> None:
    make_doc(ctx)
    ctx.run("git add .", warn=True)
    ctx.run('git commit -a -m "Update dev docs"', warn=True)
    ctx.run("git push", warn=True)


@task
def release_github(ctx) -> None:
    payload = {
        "tag_name": "v" + NEW_VER,
        "target_commitish": "master",
        "name": "v" + NEW_VER,
        "body": "See changes at https://materialsproject.github.io/custodian",
        "draft": False,
        "prerelease": False,
    }
    response = requests.post(
        "https://api.github.com/repos/materialsproject/custodian/releases",
        data=json.dumps(payload),
        headers={"Authorization": "token " + os.environ["GITHUB_RELEASES_TOKEN"]},
    )
    print(response.text)


@task
def test(ctx) -> None:
    ctx.run("pytest custodian")


@task
def set_ver(ctx) -> None:
    with open("pyproject.toml") as file:
        lines = [re.sub(r"^version = \"([^,]+)\"", f'version = "{NEW_VER}"', line.rstrip()) for line in file]

    with open("pyproject.toml", "w") as file:
        file.write("\n".join(lines) + "\n")

    ctx.run("ruff check --fix src")
    ctx.run("ruff format pyproject.toml")


@task
def update_changelog(ctx, version=None, sim=False) -> None:
    """
    Create a preliminary change log using the git logs.

    :param ctx:
    """
    version = version or datetime.datetime.now().strftime("%Y.%-m.%-d")
    output = subprocess.check_output(["git", "log", "--pretty=format:%s", f"v{CURRENT_VER}..HEAD"])
    lines = []
    misc = []
    for line in output.decode("utf-8").strip().split("\n"):
        m = re.search(r"\(\#(\d+)\)", line)
        if m:
            pr_number = m.group(1)
            pr_name = m.group().rsplit(r"\(", 1)[0]
            response = requests.get(f"https://api.github.com/repos/materialsproject/custodian/pulls/{pr_number}")
            try:
                dct = response.json()
                contrib = dct["user"]["login"]
                lines.append(f"* PR #{pr_number} from @{contrib} {pr_name}")
                if "body" in response.json():
                    for ll in response.json()["body"].split("\n"):
                        ll = ll.strip()
                        if ll in {"", "## Summary"}:
                            continue
                        if ll.startswith(("## Checklist", "## TODO")):
                            break
                        lines.append(f"    {ll}")
            except:
                pass
        else:
            misc.append("- " + line)
    with open("docs/changelog.md") as file:
        contents = file.read()
    head = "# Change Log"
    i = contents.find(head)
    i += len(head)

    contents = contents[0:i] + f"\n\n## {NEW_VER}\n" + "\n".join(lines) + contents[i:]
    if not sim:
        with open("docs/changelog.md", "w") as file:
            file.write(contents)
        ctx.run("open docs/changelog.md")
    else:
        print(contents)
    print("The following commit messages were not included...")
    print("\n".join(misc))


@task
def release(ctx) -> None:
    set_ver(ctx)
    update_doc(ctx)
    release_github(ctx)



"""Basic script to run nwchem job."""

import logging

from custodian.custodian import Custodian
from custodian.nwchem.handlers import NwchemErrorHandler
from custodian.nwchem.jobs import NwchemJob


def do_run(args) -> None:
    """Do the run."""
    logging.basicConfig(format="%(asctime)s %(message)s", level=logging.INFO, filename="run.log")
    job = NwchemJob(
        nwchem_cmd=args.command.split(),
        input_file=args.infile,
        output_file=args.outfile,
    )
    c = Custodian(
        [NwchemErrorHandler(output_filename=args.outfile)],
        [job],
        max_errors=5,
        scratch_dir=args.scratch,
        gzipped_output=args.gzipped,
        checkpoint=True,
    )
    c.run()


def main() -> None:
    """Main method."""
    import argparse

    parser = argparse.ArgumentParser(
        description="""
    run_nwchem is a master script to perform various kinds of Nwchem runs.
    """,
        epilog="""Author: Shyue Ping Ong""",
    )

    parser.add_argument(
        "-c",
        "--command",
        dest="command",
        nargs="?",
        default="nwchem",
        type=str,
        help='Nwchem command. Defaults to nwchem. If you are using mpirun, set this to something like "mpirun nwchem".',
    )

    parser.add_argument(
        "-s",
        "--scratch",
        dest="scratch",
        nargs="?",
        default=None,
        type=str,
        help="Scratch directory to perform run in. Specify the root scratch "
        "directory as the code will automatically create a temporary "
        "subdirectory to run the job.",
    )

    parser.add_argument(
        "-i",
        "--infile",
        dest="infile",
        nargs="?",
        default="mol.nw",
        type=str,
        help="Input filename.",
    )

    parser.add_argument(
        "-o",
        "--output",
        dest="outfile",
        nargs="?",
        default="mol.nwout",
        type=str,
        help="Output filename.",
    )

    parser.add_argument(
        "-z",
        "--gzip",
        dest="gzip",
        action="store_true",
        help="Add this option to gzip the final output. Do not gzip if you "
        "are going to perform an additional static run.",
    )

    args = parser.parse_args()
    do_run(args)


if __name__ == "__main__":
    main()



import datetime
import glob
import gzip
import os
import shutil
from unittest import TestCase

from custodian.gaussian.handlers import GaussianErrorHandler, WallTimeErrorHandler
from tests.conftest import TEST_FILES

__author__ = "Rasha Atwi"
__version__ = "0.1"
__maintainer__ = "Rasha Atwi"
__email__ = "rasha.atwi@stonybrook.edu"
__status__ = "Alpha"
__date__ = "3/21/24"

TEST_DIR = f"{TEST_FILES}/gaussian"
SCR_DIR = f"{TEST_DIR}/scratch"
CWD = os.getcwd()


def gunzip_file(gauss_file):
    output_file = os.path.splitext(gauss_file)[0]
    if not os.path.exists(output_file) and os.path.exists(gauss_file):
        with gzip.open(gauss_file, "rb") as f_in, open(output_file, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)
    return output_file


class TestGaussianErrorHandler(TestCase):
    def setUp(self):
        os.makedirs(SCR_DIR, exist_ok=True)
        os.chdir(SCR_DIR)

    def test_opt_steps_cycles(self):
        gunzip_file(f"{TEST_DIR}/opt_steps_cycles.out.gz")
        for file in ["opt_steps_cycles.com", "opt_steps_cycles.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="opt_steps_cycles.com",
            output_file="opt_steps_cycles.out",
            opt_max_cycles=100,
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["opt_steps"]
        assert dct["actions"] == [
            {"structure": "from_final_structure"},
            {"opt_max_cycles": 100},
        ]

    def test_opt_steps_from_structure(self):
        gunzip_file(f"{TEST_DIR}/opt_steps_from_structure.out.gz")
        for file in ["opt_steps_from_structure.com", "opt_steps_from_structure.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="opt_steps_from_structure.com",
            output_file="opt_steps_from_structure.out",
            opt_max_cycles=5,
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["opt_steps"]
        assert dct["actions"] == [{"structure": "from_final_structure"}]

    def test_opt_steps_int_grid(self):
        gunzip_file(f"{TEST_DIR}/opt_steps_int_grid.out.gz")
        for file in ["opt_steps_int_grid.com", "opt_steps_int_grid.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="opt_steps_int_grid.com",
            output_file="opt_steps_int_grid.out",
            opt_max_cycles=1,
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["opt_steps"]
        assert dct["actions"] == [{"integral": "ultra_fine"}]

    def test_opt_steps_better_guess(self):
        gunzip_file(f"{TEST_DIR}/opt_steps_better_guess.out.gz")
        for file in ["opt_steps_better_guess.com", "opt_steps_better_guess.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="opt_steps_better_guess.com",
            output_file="opt_steps_better_guess.out",
            opt_max_cycles=1,
            lower_functional="HF",
            lower_basis_set="STO-3G",
            job_type="better_guess",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["opt_steps"]
        assert dct["actions"] == [{"opt_level_of_theory": "better_geom_guess"}]

        GaussianErrorHandler.activate_better_guess = False

    def test_scf_convergence_cycles(self):
        gunzip_file(f"{TEST_DIR}/scf_convergence_cycles.out.gz")
        for file in ["scf_convergence_cycles.com", "scf_convergence_cycles.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="scf_convergence_cycles.com",
            output_file="scf_convergence_cycles.out",
            scf_max_cycles=100,
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["scf_convergence"]
        assert dct["actions"] == [{"scf_max_cycles": 100}]

    def test_scf_convergence_algorithm(self):
        gunzip_file(f"{TEST_DIR}/scf_convergence_algorithm.out.gz")
        for file in ["scf_convergence_algorithm.com", "scf_convergence_algorithm.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="scf_convergence_algorithm.com",
            output_file="scf_convergence_algorithm.out",
            scf_max_cycles=1,
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["scf_convergence"]
        assert dct["actions"] == [{"scf_algorithm": "xqc"}]

    def test_scf_convergence_better_guess(self):
        gunzip_file(f"{TEST_DIR}/scf_convergence_better_guess.out.gz")
        for file in [
            "scf_convergence_better_guess.com",
            "scf_convergence_better_guess.out",
        ]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="scf_convergence_better_guess.com",
            output_file="scf_convergence_better_guess.out",
            scf_max_cycles=3,
            lower_functional="HF",
            lower_basis_set="STO-3G",
            job_type="better_guess",
        )
        handler.activate_better_guess = False
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["scf_convergence"]
        assert dct["actions"] == [{"scf_level_of_theory": "better_scf_guess"}]

        GaussianErrorHandler.activate_better_guess = False

    def test_linear_bend(self):
        gunzip_file(f"{TEST_DIR}/linear_bend.out.gz")
        for file in ["linear_bend.com", "linear_bend.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="linear_bend.com",
            output_file="linear_bend.out",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["linear_bend"]
        assert dct["actions"] == [{"coords": "rebuild_redundant_internals"}]

    def test_solute_solvent_surface(self):
        gunzip_file(f"{TEST_DIR}/solute_solvent_surface.out.gz")
        for file in ["solute_solvent_surface.com", "solute_solvent_surface.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="solute_solvent_surface.com",
            output_file="solute_solvent_surface.out",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["solute_solvent_surface"]
        assert dct["actions"] == [{"surface": "SAS"}]

    def test_internal_coords(self):
        pass

    def test_blank_line(self):
        gunzip_file(f"{TEST_DIR}/zmatrix.out.gz")
        for file in ["zmatrix.com", "zmatrix.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="zmatrix.com",
            output_file="zmatrix.out",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["zmatrix"]
        assert dct["actions"] == [{"blank_lines": "rewrite_input_file"}]

    def test_missing_mol(self):
        gunzip_file(f"{TEST_DIR}/missing_mol.out.gz")
        for file in ["missing_mol.com", "missing_mol.out", "Optimization.chk"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="missing_mol.com",
            output_file="missing_mol.out",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["missing_mol"]
        assert dct["actions"] == [{"mol": "get_from_checkpoint"}]

    def test_found_coords(self):
        gunzip_file(f"{TEST_DIR}/found_coords.out.gz")
        for file in ["found_coords.com", "found_coords.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="found_coords.com",
            output_file="found_coords.out",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["found_coords"]
        assert dct["actions"] == [{"mol": "remove_from_input"}]

    def test_coords_dict_geom(self):
        gunzip_file(f"{TEST_DIR}/coords_dict_geom.out.gz")
        for file in ["coords_dict_geom.com", "coords_dict_geom.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="coords_dict_geom.com",
            output_file="coords_dict_geom.out",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["coords"]
        assert dct["actions"] == [{"coords": "remove_connectivity"}]

    def test_coords_string_geom(self):
        gunzip_file(f"{TEST_DIR}/coords_string_geom.out.gz")
        for file in ["coords_string_geom.com", "coords_string_geom.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="coords_string_geom.com",
            output_file="coords_string_geom.out",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["coords"]
        assert dct["actions"] == [{"coords": "remove_connectivity"}]

    def test_missing_file(self):
        gunzip_file(f"{TEST_DIR}/missing_file.out.gz")
        for file in ["missing_file.com", "missing_file.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="missing_file.com",
            output_file="missing_file.out",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["missing_file"]
        assert dct["actions"] is None

    def test_bad_file(self):
        gunzip_file(f"{TEST_DIR}/bad_file.out.gz")
        for file in ["bad_file.com", "bad_file.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="bad_file.com",
            output_file="bad_file.out",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["bad_file"]
        assert dct["actions"] is None

    def test_coord_inputs(self):
        gunzip_file(f"{TEST_DIR}/coord_inputs.out.gz")
        for file in ["coord_inputs.com", "coord_inputs.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="coord_inputs.com",
            output_file="coord_inputs.out",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["coord_inputs"]
        assert dct["actions"] == [{"coords": "use_zmatrix_format"}]

    def test_syntax(self):
        gunzip_file(f"{TEST_DIR}/syntax.out.gz")
        for file in ["syntax.com", "syntax.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="syntax.com",
            output_file="syntax.out",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["syntax"]
        assert dct["actions"] is None

    def test_insufficient_memory(self):
        gunzip_file(f"{TEST_DIR}/insufficient_memory.out.gz")
        for file in ["insufficient_memory.com", "insufficient_memory.out"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")
        handler = GaussianErrorHandler(
            input_file="insufficient_memory.com",
            output_file="insufficient_memory.out",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["insufficient_mem"]
        assert dct["actions"] == [{"memory": "increase_to_gaussian_recommendation"}]

    def tearDown(self):
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)
        files_to_remove = glob.glob(f"{TEST_DIR}/*.out")
        if files_to_remove and glob.glob(f"{TEST_DIR}/*.out.gz"):
            for file_path in files_to_remove:
                os.remove(file_path)


class TestWallTimeErrorHandler(TestCase):
    def setUp(self):
        os.makedirs(SCR_DIR, exist_ok=True)
        os.chdir(SCR_DIR)
        os.environ.pop("JOB_START_TIME", None)
        gunzip_file(f"{TEST_DIR}/walltime.out.gz")
        for file in ["walltime.com", "walltime.out", "Gau-mock.rwf"]:
            shutil.copyfile(f"{TEST_DIR}/{file}", f"{SCR_DIR}/{file}")

    def test_walltime_init(self):
        handler = WallTimeErrorHandler(
            wall_time=3600,
            buffer_time=300,
            input_file="wall_time.com",
            output_file="wall_time.out",
        )
        init_time = handler.init_time
        assert os.environ.get("JOB_START_TIME") == f"{init_time:%a %b %d %H:%M:%S UTC %Y}"
        # Test that walltime persists if new handler is created
        handler = WallTimeErrorHandler(
            wall_time=3600,
            buffer_time=300,
            input_file="walltime.com",
            output_file="walltime.out",
        )
        assert os.environ.get("JOB_START_TIME") == f"{init_time:%a %b %d %H:%M:%S UTC %Y}"

    def test_walltime_check_and_correct(self):
        # Try a 1 hr wall time with a 5 mins buffer
        handler = WallTimeErrorHandler(
            wall_time=3600,
            buffer_time=300,
            input_file="walltime.com",
            output_file="walltime.out",
        )
        assert not handler.check()

        # Make sure the check returns True when the remaining time is <= buffer time
        handler.init_time = datetime.datetime.now() - datetime.timedelta(minutes=59)
        assert handler.check()

        # Test that the input file is written correctly
        handler.correct()
        assert os.path.exists("walltime.com.wt")
        with open("walltime.com.wt") as file:
            first_line = file.readline().strip()
        # assert first_line == "%rwf=./Gau-mock.rwf"
        assert "rwf" in first_line

    def tearDown(self):
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)
        files_to_remove = glob.glob(f"{TEST_DIR}/*.out")
        if files_to_remove and glob.glob(f"{TEST_DIR}/*.out.gz"):
            for file_path in files_to_remove:
                os.remove(file_path)



import os
import shutil
import unittest
from glob import glob

from custodian.feff.handlers import UnconvergedErrorHandler
from tests.conftest import TEST_FILES

__author__ = "Chen Zheng"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Chen Zheng"
__email__ = "chz022@ucsd.edu"
__date__ = "Oct 18, 2017"


def clean_dir() -> None:
    for file in glob("error.*.tar.gz"):
        os.remove(file)


class UnconvergedErrorHandlerTest(unittest.TestCase):
    def setUp(self) -> None:
        os.chdir(TEST_FILES)
        subdir = f"{TEST_FILES}/feff_unconverged"
        os.chdir(subdir)
        shutil.copy("ATOMS", "ATOMS.orig")
        shutil.copy("PARAMETERS", "PARAMETERS.orig")
        shutil.copy("HEADER", "HEADER.orig")
        shutil.copy("POTENTIALS", "POTENTIALS.orig")
        shutil.copy("feff.inp", "feff.inp.orig")
        shutil.copy("log1.dat", "log1.dat.orig")

    def test_check_unconverged(self) -> None:
        handler = UnconvergedErrorHandler()
        assert handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["Non-converging job"]
        assert dct["actions"] == [
            {"dict": "PARAMETERS", "action": {"_set": {"RESTART": []}}},
            {"action": {"_set": {"SCF": [7, 0, 100, 0.2, 3]}}, "dict": "PARAMETERS"},
        ]
        shutil.move("ATOMS.orig", "ATOMS")
        shutil.move("PARAMETERS.orig", "PARAMETERS")
        shutil.move("HEADER.orig", "HEADER")
        shutil.move("POTENTIALS.orig", "POTENTIALS")
        shutil.move("feff.inp.orig", "feff.inp")
        shutil.move("log1.dat.orig", "log1.dat")
        clean_dir()



import os
import shutil

import pytest

from custodian.vasp.validators import VaspAECCARValidator, VaspFilesValidator, VaspNpTMDValidator, VasprunXMLValidator
from tests.conftest import TEST_FILES


@pytest.fixture(autouse=True)
def _clear_tracked_cache() -> None:
    """Clear the cache of the stored functions between the tests."""
    from custodian.utils import tracked_lru_cache

    tracked_lru_cache.tracked_cache_clear()


class TestVasprunXMLValidator:
    def test_check_and_correct(self) -> None:
        os.chdir(f"{TEST_FILES}/bad_vasprun")
        handler = VasprunXMLValidator()
        assert handler.check()

        # Unconverged still has a valid vasprun.
        os.chdir(f"{TEST_FILES}/unconverged")
        shutil.copy("vasprun.xml.electronic", "vasprun.xml")
        assert not handler.check()
        os.remove("vasprun.xml")

    def test_as_dict(self) -> None:
        handler = VasprunXMLValidator()
        dct = handler.as_dict()
        h2 = VasprunXMLValidator.from_dict(dct)
        assert isinstance(h2, VasprunXMLValidator)


class TestVaspFilesValidator:
    def test_check_and_correct(self) -> None:
        # just an example where CONTCAR is not present
        os.chdir(f"{TEST_FILES}/positive_energy")
        handler = VaspFilesValidator()
        assert handler.check()

        os.chdir(f"{TEST_FILES}/postprocess")
        assert not handler.check()

    def test_as_dict(self) -> None:
        handler = VaspFilesValidator()
        dct = handler.as_dict()
        h2 = VaspFilesValidator.from_dict(dct)
        assert isinstance(h2, VaspFilesValidator)


class TestVaspNpTMDValidator:
    def test_check_and_correct(self) -> None:
        # NPT-AIMD using correct VASP
        os.chdir(f"{TEST_FILES}/npt_common")
        handler = VaspNpTMDValidator()
        assert not handler.check()

        # NVT-AIMD using correct VASP
        os.chdir(f"{TEST_FILES}/npt_nvt")
        assert not handler.check()

        # NPT-AIMD using incorrect VASP
        os.chdir(f"{TEST_FILES}/npt_bad_vasp")
        assert handler.check()

    def test_as_dict(self) -> None:
        handler = VaspNpTMDValidator()
        dct = handler.as_dict()
        h2 = VaspNpTMDValidator.from_dict(dct)
        assert isinstance(h2, VaspNpTMDValidator)


class TestVaspAECCARValidator:
    def test_check_and_correct(self) -> None:
        # NPT-AIMD using correct VASP
        os.chdir(f"{TEST_FILES}/bad_aeccar")
        handler = VaspAECCARValidator()
        assert handler.check()



import os
import shutil
import unittest
from unittest import TestCase

from pymatgen.io.qchem.inputs import QCInput

from custodian.qchem.handlers import QChemErrorHandler
from tests.conftest import TEST_FILES

try:
    from openbabel import openbabel as ob
except ImportError:
    ob = None

__author__ = "Samuel Blau, Brandon Woods, Shyam Dwaraknath, Ryan Kingsbury"
__copyright__ = "Copyright 2018, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Samuel Blau"
__email__ = "samblau1@gmail.com"
__status__ = "Alpha"
__date__ = "3/26/18"
__credits__ = "Xiaohui Qu"


TEST_DIR = f"{TEST_FILES}/qchem/new_test_files"
SCR_DIR = f"{TEST_DIR}/scratch"
CWD = os.getcwd()
skip_if_no_openbabel = unittest.skipIf(ob is None, "openbabel not installed")


@skip_if_no_openbabel
class QChemErrorHandlerTest(TestCase):
    def setUp(self) -> None:
        os.makedirs(SCR_DIR)
        os.chdir(SCR_DIR)

    def _check_equivalent_inputs(self, input1, input2) -> None:
        QCinput1 = QCInput.from_file(input1)
        QCinput2 = QCInput.from_file(input2)
        sections1 = QCInput.find_sections(QCinput1.get_str())
        sections2 = QCInput.find_sections(QCinput2.get_str())
        assert sections1 == sections2
        for key in sections1:
            assert QCinput1.as_dict().get(key) == QCinput2.as_dict().get(key)

    def test_unable_to_determine_lamda(self) -> None:
        for ii in range(2):
            shutil.copyfile(
                f"{TEST_DIR}/unable_to_determine_lamda.qin.{ii}",
                f"{SCR_DIR}/unable_to_determine_lamda.qin.{ii}",
            )
            shutil.copyfile(
                f"{TEST_DIR}/unable_to_determine_lamda.qout.{ii}",
                f"{SCR_DIR}/unable_to_determine_lamda.qout.{ii}",
            )

        handler = QChemErrorHandler(
            input_file="unable_to_determine_lamda.qin.0",
            output_file="unable_to_determine_lamda.qout.0",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["unable_to_determine_lamda"]
        assert dct["actions"] == [
            {"s2thresh": "16"},
            {"molecule": "molecule_from_last_geometry"},
            {"scf_algorithm": "gdm"},
            {"max_scf_cycles": "500"},
        ]
        self._check_equivalent_inputs("unable_to_determine_lamda.qin.0", "unable_to_determine_lamda.qin.1")

    def test_linear_dependent_basis_and_FileMan(self) -> None:
        for ii in range(1, 3):
            shutil.copyfile(
                f"{TEST_DIR}/unable_to_determine_lamda.qin.{ii}",
                f"{SCR_DIR}/unable_to_determine_lamda.qin.{ii}",
            )
            shutil.copyfile(
                f"{TEST_DIR}/unable_to_determine_lamda.qout.{ii}",
                f"{SCR_DIR}/unable_to_determine_lamda.qout.{ii}",
            )

        handler = QChemErrorHandler(
            input_file="unable_to_determine_lamda.qin.1",
            output_file="unable_to_determine_lamda.qout.1",
        )
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["premature_end_FileMan_error"]
        assert dct["warnings"]["linear_dependence"] is True
        assert dct["actions"] == [{"scf_guess_always": "true"}]

    def test_failed_to_transform(self) -> None:
        for ii in range(2):
            shutil.copyfile(f"{TEST_DIR}/qunino_vinyl.qin.{ii}", f"{SCR_DIR}/qunino_vinyl.qin.{ii}")
            shutil.copyfile(f"{TEST_DIR}/qunino_vinyl.qout.{ii}", f"{SCR_DIR}/qunino_vinyl.qout.{ii}")

        handler = QChemErrorHandler(input_file="qunino_vinyl.qin.0", output_file="qunino_vinyl.qout.0")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["failed_to_transform_coords"]
        assert dct["actions"] == [{"thresh": "14"}, {"s2thresh": "16"}, {"sym_ignore": "true"}, {"symmetry": "false"}]
        self._check_equivalent_inputs("qunino_vinyl.qin.0", "qunino_vinyl.qin.1")

        handler = QChemErrorHandler(input_file="qunino_vinyl.qin.1", output_file="qunino_vinyl.qout.1")
        assert handler.check() is False

    def test_scf_failed_to_converge(self) -> None:
        for ii in range(3):
            shutil.copyfile(f"{TEST_DIR}/crowd_gradient.qin.{ii}", f"{SCR_DIR}/crowd_gradient.qin.{ii}")
            shutil.copyfile(f"{TEST_DIR}/crowd_gradient.qout.{ii}", f"{SCR_DIR}/crowd_gradient.qout.{ii}")

        handler = QChemErrorHandler(input_file="crowd_gradient.qin.0", output_file="crowd_gradient.qout.0")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["SCF_failed_to_converge"]
        assert dct["actions"] == [{"s2thresh": "16"}, {"max_scf_cycles": 100}, {"thresh": "14"}]
        self._check_equivalent_inputs("crowd_gradient.qin.0", "crowd_gradient.qin.1")

    def test_scf_failed_to_converge_gdm_add_cycles(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/gdm_add_cycles/mol.qin", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/gdm_add_cycles/mol.qin.1", f"{SCR_DIR}/mol.qin.1")
        shutil.copyfile(f"{TEST_DIR}/gdm_add_cycles/mol.qout", f"{SCR_DIR}/mol.qout")

        handler = QChemErrorHandler(input_file="mol.qin", output_file="mol.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["SCF_failed_to_converge"]
        assert dct["actions"] == [{"max_scf_cycles": "500"}]
        self._check_equivalent_inputs("mol.qin", "mol.qin.1")

    def test_advanced_scf_failed_to_converge_1(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/diis_guess_always/mol.qin.0", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/diis_guess_always/mol.qout.0", f"{SCR_DIR}/mol.qout")
        shutil.copyfile(f"{TEST_DIR}/diis_guess_always/mol.qin.1", f"{SCR_DIR}/mol.qin.1")

        handler = QChemErrorHandler(input_file="mol.qin", output_file="mol.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["SCF_failed_to_converge"]
        assert dct["actions"] == [{"scf_algorithm": "gdm"}, {"max_scf_cycles": "500"}]
        self._check_equivalent_inputs("mol.qin", "mol.qin.1")

    def test_scf_into_opt(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/scf_into_opt/mol.qin.0", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/scf_into_opt/mol.qout.0", f"{SCR_DIR}/mol.qout")
        shutil.copyfile(f"{TEST_DIR}/scf_into_opt/mol.qin.1", f"{SCR_DIR}/mol.qin.1")

        handler = QChemErrorHandler(input_file="mol.qin", output_file="mol.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["SCF_failed_to_converge"]
        assert dct["actions"] == [{"scf_algorithm": "gdm"}, {"max_scf_cycles": "500"}]
        self._check_equivalent_inputs("mol.qin", "mol.qin.1")

        shutil.copyfile(f"{TEST_DIR}/scf_into_opt/mol.qout.1", f"{SCR_DIR}/mol.qout")

        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["out_of_opt_cycles"]
        assert dct["actions"] == [{"molecule": "molecule_from_last_geometry"}]

    def test_custom_smd(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/custom_smd/mol.qin.0", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/custom_smd/mol.qout.0", f"{SCR_DIR}/mol.qout")
        shutil.copyfile(f"{TEST_DIR}/custom_smd/mol.qin.1", f"{SCR_DIR}/mol.qin.1")

        handler = QChemErrorHandler(input_file="mol.qin", output_file="mol.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["SCF_failed_to_converge"]
        assert dct["actions"] == [{"scf_algorithm": "gdm"}, {"max_scf_cycles": "500"}]
        self._check_equivalent_inputs("mol.qin", "mol.qin.1")

        shutil.copyfile(f"{TEST_DIR}/custom_smd/mol.qout.1", f"{SCR_DIR}/mol.qout")

        handler.check()
        dct = handler.correct()
        assert dct["errors"] == []
        assert dct["actions"] is None

    def test_out_of_opt_cycles(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/crowd_gradient.qin.2", f"{SCR_DIR}/crowd_gradient.qin.2")
        shutil.copyfile(f"{TEST_DIR}/crowd_gradient.qout.2", f"{SCR_DIR}/crowd_gradient.qout.2")
        shutil.copyfile(f"{TEST_DIR}/crowd_gradient.qin.3", f"{SCR_DIR}/crowd_gradient.qin.3")

        handler = QChemErrorHandler(input_file="crowd_gradient.qin.2", output_file="crowd_gradient.qout.2")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["out_of_opt_cycles"]
        assert dct["actions"] == [{"geom_max_cycles:": 200}, {"molecule": "molecule_from_last_geometry"}]
        self._check_equivalent_inputs("crowd_gradient.qin.2", "crowd_gradient.qin.3")

    def test_advanced_out_of_opt_cycles(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/2564_complete/error1/mol.qin", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/2564_complete/error1/mol.qout", f"{SCR_DIR}/mol.qout")
        shutil.copyfile(f"{TEST_DIR}/2564_complete/mol.qin.opt_0", f"{SCR_DIR}/mol.qin.opt_0")
        handler = QChemErrorHandler(input_file="mol.qin", output_file="mol.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["out_of_opt_cycles"]
        assert dct["actions"] == [{"s2thresh": "16"}, {"molecule": "molecule_from_last_geometry"}]
        self._check_equivalent_inputs("mol.qin.opt_0", "mol.qin")
        assert handler.opt_error_history[0] == "more_bonds"
        shutil.copyfile(f"{TEST_DIR}/2564_complete/mol.qin.opt_0", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/2564_complete/mol.qout.opt_0", f"{SCR_DIR}/mol.qout")
        handler.check()
        assert handler.opt_error_history == []

    def test_advanced_out_of_opt_cycles1(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/2620_complete/mol.qin.opt_0", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/2620_complete/mol.qout.opt_0", f"{SCR_DIR}/mol.qout")
        handler = QChemErrorHandler(input_file="mol.qin", output_file="mol.qout")
        assert handler.check() is False

    def test_failed_to_read_input(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/unable_lamda_weird.qin", f"{SCR_DIR}/unable_lamda_weird.qin")
        shutil.copyfile(f"{TEST_DIR}/unable_lamda_weird.qout", f"{SCR_DIR}/unable_lamda_weird.qout")
        handler = QChemErrorHandler(input_file="unable_lamda_weird.qin", output_file="unable_lamda_weird.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["failed_to_read_input"]
        assert dct["actions"] == [{"rerun_job_no_changes": True}]
        self._check_equivalent_inputs("unable_lamda_weird.qin.last", "unable_lamda_weird.qin")

    def test_input_file_error(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/bad_input.qin", f"{SCR_DIR}/bad_input.qin")
        shutil.copyfile(f"{TEST_DIR}/bad_input.qout", f"{SCR_DIR}/bad_input.qout")
        handler = QChemErrorHandler(input_file="bad_input.qin", output_file="bad_input.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["input_file_error"]
        assert dct["actions"] is None

    def test_basis_not_supported(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/basis_not_supported.qin", f"{SCR_DIR}/basis_not_supported.qin")
        shutil.copyfile(f"{TEST_DIR}/basis_not_supported.qout", f"{SCR_DIR}/basis_not_supported.qout")
        handler = QChemErrorHandler(input_file="basis_not_supported.qin", output_file="basis_not_supported.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["basis_not_supported"]
        assert dct["actions"] is None

    def test_NLebdevPts(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/lebdevpts.qin", f"{SCR_DIR}/lebdevpts.qin")
        shutil.copyfile(f"{TEST_DIR}/lebdevpts.qout", f"{SCR_DIR}/lebdevpts.qout")
        handler = QChemErrorHandler(input_file="lebdevpts.qin", output_file="lebdevpts.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["NLebdevPts"]
        assert dct["actions"] == [{"esp_surface_density": "250"}]

    def test_read_error(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/molecule_read_error/mol.qin", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/molecule_read_error/mol.qout", f"{SCR_DIR}/mol.qout")
        handler = QChemErrorHandler(input_file="mol.qin", output_file="mol.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["read_molecule_error"]
        assert dct["actions"] == [{"rerun_job_no_changes": True}]
        self._check_equivalent_inputs("mol.qin.last", "mol.qin")

    def test_never_called_qchem_error(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/mpi_error/mol.qin", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/mpi_error/mol.qout", f"{SCR_DIR}/mol.qout")
        handler = QChemErrorHandler(input_file="mol.qin", output_file="mol.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["never_called_qchem"]
        assert dct["actions"] == [{"rerun_job_no_changes": True}]
        self._check_equivalent_inputs("mol.qin.last", "mol.qin")

    def test_OOS_read_hess(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/OOS_read_hess.qin", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/OOS_read_hess.qout", f"{SCR_DIR}/mol.qout")
        handler = QChemErrorHandler(input_file="mol.qin", output_file="mol.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["out_of_opt_cycles"]
        assert dct["actions"] == [
            {"s2thresh": "16"},
            {"molecule": "molecule_from_last_geometry"},
            {"geom_opt_hessian": "deleted"},
        ]
        self._check_equivalent_inputs(f"{TEST_DIR}/OOS_read_hess_next.qin", "mol.qin")

    def test_gdm_neg_precon_error(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/gdm_neg_precon_error.qin", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/gdm_neg_precon_error.qout", f"{SCR_DIR}/mol.qout")
        handler = QChemErrorHandler(input_file="mol.qin", output_file="mol.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["gdm_neg_precon_error"]
        assert dct["actions"] == [{"molecule": "molecule_from_last_geometry"}]

    def test_fileman_cpscf_nseg_error(self) -> None:
        shutil.copyfile(f"{TEST_DIR}/fileman_cpscf.qin", f"{SCR_DIR}/mol.qin")
        shutil.copyfile(f"{TEST_DIR}/fileman_cpscf.qout", f"{SCR_DIR}/mol.qout")
        handler = QChemErrorHandler(input_file="mol.qin", output_file="mol.qout")
        handler.check()
        dct = handler.correct()
        assert dct["errors"] == ["premature_end_FileMan_error"]
        assert dct["actions"] == [{"cpscf_nseg": "3"}]

    def tearDown(self) -> None:
        os.chdir(CWD)
        shutil.rmtree(SCR_DIR)



import os

from custodian.lobster.handlers import ChargeSpillingValidator, EnoughBandsValidator, LobsterFilesValidator
from tests.conftest import TEST_FILES

test_files_lobster = f"{TEST_FILES}/lobster/lobsterouts"


class TestChargeSpillingValidator:
    def test_check_and_correct(self) -> None:
        v = ChargeSpillingValidator(output_filename=f"{test_files_lobster}/lobsterout.normal")
        assert not v.check()

        v2 = ChargeSpillingValidator(output_filename=f"{test_files_lobster}/lobsterout.largespilling")
        assert v2.check()

        v2b = ChargeSpillingValidator(output_filename=f"{test_files_lobster}/lobsterout.largespilling_2")
        assert v2b.check()

        v3 = ChargeSpillingValidator(output_filename=f"{test_files_lobster}/nolobsterout/lobsterout")
        assert not v3.check()

        v4 = ChargeSpillingValidator(output_filename=f"{test_files_lobster}/no_spin/lobsterout")
        assert not v4.check()

    def test_as_dict(self) -> None:
        v = ChargeSpillingValidator(output_filename=f"{test_files_lobster}/lobsterout.normal")
        dct = v.as_dict()
        v2 = ChargeSpillingValidator.from_dict(dct)
        assert isinstance(v2, ChargeSpillingValidator)


class TestLobsterFilesValidator:
    def test_check_and_correct_1(self) -> None:
        os.chdir(test_files_lobster)
        v = LobsterFilesValidator()
        assert not v.check()

    def test_check_and_correct_2(self) -> None:
        os.chdir(f"{test_files_lobster}/../lobsterins")
        v2 = LobsterFilesValidator()
        assert v2.check()

    def test_check_and_correct_3(self) -> None:
        os.chdir(f"{test_files_lobster}/crash")
        v3 = LobsterFilesValidator()
        assert v3.check()

    def test_as_dict(self) -> None:
        os.chdir(test_files_lobster)
        v = LobsterFilesValidator()
        dct = v.as_dict()
        v2 = LobsterFilesValidator.from_dict(dct)
        assert isinstance(v2, LobsterFilesValidator)


class TestEnoughBandsValidator:
    def test_check_and_correct(self) -> None:
        v = EnoughBandsValidator(output_filename=f"{test_files_lobster}/lobsterout.normal")
        assert not v.check()

    def test_check_and_correct2(self) -> None:
        v2 = EnoughBandsValidator(output_filename=f"{test_files_lobster}/lobsterout.nocohp")
        assert v2.check()

    def test_check_and_correct3(self) -> None:
        v3 = EnoughBandsValidator(output_filename=f"{test_files_lobster}/nolobsterout/lobsterout")
        assert not v3.check()

    def test_as_dict(self) -> None:
        v = EnoughBandsValidator(output_filename=f"{test_files_lobster}/lobsterout.normal")
        dct = v.as_dict()
        v2 = EnoughBandsValidator.from_dict(dct)
        assert isinstance(v2, EnoughBandsValidator)



import os
from pathlib import Path
from unittest import mock
from unittest.mock import ANY, MagicMock, patch

import psutil

from custodian.jdftx.jobs import JDFTxJob

TEST_DIR = Path(__file__).resolve().parent.parent
TEST_FILES = f"{TEST_DIR}/files/jdftx"


def create_mock_process(
    name="jdftx", open_files=None, pid=12345, name_side_effect=None, wait_side_effect=None, terminate_side_effect=None
):
    if open_files is None:  # Set a default value if not provided
        open_files = [MagicMock(path=os.path.join("default_path", "jdftx.out"))]

    mock_process = mock.Mock(spec=psutil.Process)
    mock_process.name.return_value = name
    mock_process.open_files.return_value = open_files
    mock_process.pid = pid
    mock_process.name.side_effect = name_side_effect
    mock_process.wait.side_effect = wait_side_effect
    mock_process.terminate.side_effect = terminate_side_effect
    return mock_process


def test_jdftx_job_init(jdftx_job):
    assert jdftx_job.jdftx_cmd == "jdftx"
    assert jdftx_job.input_file == "init.in"
    assert jdftx_job.output_file == "jdftx.out"
    assert jdftx_job.stderr_file == "std_err.txt"


def test_jdftx_job_setup(jdftx_job, tmp_path):
    jdftx_job.setup(str(tmp_path))
    # Setup method doesn't do anything, so just checking that it doesn't raise an exception


def test_jdftx_job_run(jdftx_job, tmp_path):
    with patch("subprocess.run") as mock_run:
        mock_process = MagicMock()
        mock_run.return_value = mock_process

        result = jdftx_job.run(str(tmp_path))

        assert result == mock_process
        mock_run.assert_called_once_with(
            ["jdftx", "-i", "init.in", "-o", "jdftx.out"],
            cwd=str(tmp_path),
            stdout=ANY,
            stderr=ANY,
            shell=False,
            check=False,
        )


def test_jdftx_job_run_creates_output_files(jdftx_job, tmp_path):
    with patch("subprocess.run"):
        jdftx_job.run(str(tmp_path))

    assert os.path.exists(os.path.join(str(tmp_path), "jdftx.out"))
    assert os.path.exists(os.path.join(str(tmp_path), "std_err.txt"))


def test_jdftx_job_postprocess(jdftx_job, tmp_path):
    jdftx_job.postprocess(str(tmp_path))
    # Postprocess method doesn't do anything, so we just check that it doesn't raise an exception


@mock.patch("psutil.pid_exists")
@mock.patch("subprocess.run")
@mock.patch.object(JDFTxJob, "terminate_process", autospec=True)
def test_jdftx_job_terminate(mock_terminate_process, mock_subprocess_run, mock_pid_exists, jdftx_job, tmp_path, caplog):
    open_files = [MagicMock(path=os.path.join(str(tmp_path), jdftx_job.output_file))]
    # Test when JDFTx process exists
    mock_process = create_mock_process(name="jdftx", open_files=open_files, pid=12345)

    with patch("psutil.process_iter", return_value=[mock_process]):
        mock_pid_exists.return_value = True
        jdftx_job.terminate(str(tmp_path))
        mock_terminate_process.assert_called_once_with(mock_process)
        mock_subprocess_run.assert_not_called()

    mock_terminate_process.reset_mock()
    mock_subprocess_run.reset_mock()

    # Test when no JDFTx process exists
    mock_process = create_mock_process(name="vasp", open_files=open_files, pid=12345)

    with patch("psutil.process_iter", return_value=[mock_process]):
        jdftx_job.terminate(str(tmp_path))
        mock_terminate_process.assert_not_called()
        mock_subprocess_run.assert_called_once_with(["killall", "jdftx"], check=False)

    mock_terminate_process.reset_mock()
    mock_subprocess_run.reset_mock()

    # Test when psutil.process_iter raises NoSuchProcess
    mock_process = create_mock_process(
        name="jdftx", open_files=open_files, pid=12345, name_side_effect=psutil.NoSuchProcess(pid=12345)
    )

    with caplog.at_level("WARNING"):
        with patch("psutil.process_iter", return_value=[mock_process]):
            jdftx_job.terminate(str(tmp_path))
            mock_terminate_process.assert_not_called()
            mock_subprocess_run.assert_called_with(["killall", "jdftx"], check=False)

        assert "Exception" in caplog.text
        assert "encountered while killing JDFTx" in caplog.text

    mock_terminate_process.reset_mock()
    mock_subprocess_run.reset_mock()

    # Test when psutil.process_iter raises AccessDenied
    with caplog.at_level("WARNING"):
        mock_process = create_mock_process(
            name="jdftx", open_files=open_files, pid=12345, name_side_effect=psutil.AccessDenied(pid=12345)
        )
        with patch("psutil.process_iter", return_value=[mock_process]):
            jdftx_job.terminate(str(tmp_path))
            mock_terminate_process.assert_not_called()
            mock_subprocess_run.assert_called_with(["killall", "jdftx"], check=False)

        assert "Exception" in caplog.text
        assert "encountered while killing JDFTx" in caplog.text


def test_terminate_process(jdftx_job, caplog):
    # Test successful termination
    mock_process = create_mock_process()
    mock_process.terminate.return_value = None  # Simulate successful termination
    mock_process.wait.return_value = None  # Simulate process finished immediately
    with caplog.at_level("INFO"):
        jdftx_job.terminate_process(mock_process)

    mock_process.terminate.assert_called_once()
    mock_process.wait.assert_called_once()

    assert "Process" in caplog.text
    assert "terminated gracefully" in caplog.text

    mock_process.reset_mock()

    # Test when process doesn't terminate gracefully
    mock_process = create_mock_process(pid=12345, wait_side_effect=psutil.TimeoutExpired(seconds=5))
    mock_process.terminate.return_value = None

    jdftx_job.terminate_process(mock_process)
    mock_process.terminate.assert_called_once()
    mock_process.kill.assert_called_once()
    mock_process.wait.assert_called()

    mock_process.reset_mock()

    # Test when process raises NoSuchProcess
    mock_process = create_mock_process(pid=12345, terminate_side_effect=psutil.NoSuchProcess(pid=12345))
    with caplog.at_level("WARNING"):
        jdftx_job.terminate_process(mock_process)

    assert "Error while terminating process" in caplog.text

    mock_process.reset_mock()

    # Test when process raises AccessDenied
    mock_process = create_mock_process(pid=12345, terminate_side_effect=psutil.AccessDenied(pid=12345))

    with caplog.at_level("WARNING"):
        jdftx_job.terminate_process(mock_process)

    assert "Error while terminating process" in caplog.text



"""Created on Jun 1, 2012."""

from __future__ import annotations

from typing import Any

import pytest

from custodian.ansible.actions import FileActions
from custodian.ansible.interpreter import Modder

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "shyue@mit.edu"
__date__ = "Jun 1, 2012"


class TestModder:
    def test_dict_modify(self) -> None:
        modder = Modder()
        dct = {"Hello": "World"}
        mod: dict[str, Any] = {"_set": {"Hello": "Universe", "Bye": "World"}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World", "Hello": "Universe"}
        mod = {"_unset": {"Hello": 1}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World"}
        mod = {"_push": {"List": 1}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World", "List": [1]}
        mod = {"_push": {"List": 2}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World", "List": [1, 2]}
        mod = {"_inc": {"num": 5}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World", "List": [1, 2], "num": 5}
        mod = {"_inc": {"num": 5}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World", "List": [1, 2], "num": 10}
        mod = {"_rename": {"num": "number"}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World", "List": [1, 2], "number": 10}
        mod = {"_add_to_set": {"List": 2}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World", "List": [1, 2], "number": 10}
        mod = {"_add_to_set": {"List": 3}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World", "List": [1, 2, 3], "number": 10}
        mod = {"_add_to_set": {"number": 3}}
        with pytest.raises(ValueError, match="Keyword number does not refer to an array"):
            modder.modify(mod, dct)
        mod = {"_pull": {"List": 1}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World", "List": [2, 3], "number": 10}
        mod = {"_pull_all": {"List": [2, 3]}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World", "List": [], "number": 10}
        mod = {"_push_all": {"List": list(range(10))}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World", "List": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9], "number": 10}
        mod = {"_pop": {"List": 1}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World", "List": [0, 1, 2, 3, 4, 5, 6, 7, 8], "number": 10}
        mod = {"_pop": {"List": -1}}
        modder.modify(mod, dct)
        assert dct == {"Bye": "World", "List": [1, 2, 3, 4, 5, 6, 7, 8], "number": 10}
        dct = {}
        mod = {"_set": {"a->b->c": 100}}
        modder.modify(mod, dct)
        assert dct == {"a": {"b": {"c": 100}}}
        mod = {"_set": {"a->b->dct": 200}}
        modder.modify(mod, dct)
        assert dct == {"a": {"b": {"c": 100, "dct": 200}}}
        mod = {"_set": {"a->b->dct": 300}}
        modder.modify(mod, dct)
        assert dct == {"a": {"b": {"c": 100, "dct": 300}}}
        mod = {"_unset": {"a->b->dct": 300}}
        modder.modify(mod, dct)
        assert dct == {"a": {"b": {"c": 100}}}
        mod = {"_push": {"a->e->f": 300}}
        modder.modify(mod, dct)
        assert dct == {"a": {"b": {"c": 100}, "e": {"f": [300]}}}
        mod = {"_push_all": {"a->e->f": [100, 200]}}
        modder.modify(mod, dct)
        assert dct == {"a": {"b": {"c": 100}, "e": {"f": [300, 100, 200]}}}
        mod = {"_inc": {"a->b->c": 2}}
        modder.modify(mod, dct)
        assert dct == {"a": {"b": {"c": 102}, "e": {"f": [300, 100, 200]}}}
        mod = {"_pull": {"a->e->f": 300}}
        modder.modify(mod, dct)
        assert dct == {"a": {"b": {"c": 102}, "e": {"f": [100, 200]}}}
        mod = {"_pull_all": {"a->e->f": [100, 200]}}
        modder.modify(mod, dct)
        assert dct == {"a": {"b": {"c": 102}, "e": {"f": []}}}
        mod = {"_push_all": {"a->e->f": [101, 201, 301, 401]}}
        modder.modify(mod, dct)
        assert dct == {"a": {"b": {"c": 102}, "e": {"f": [101, 201, 301, 401]}}}
        mod = {"_pop": {"a->e->f": 1}}
        modder.modify(mod, dct)
        assert dct == {"a": {"b": {"c": 102}, "e": {"f": [101, 201, 301]}}}
        mod = {"_pop": {"a->e->f": -1}}
        modder.modify(mod, dct)
        assert dct == {"a": {"b": {"c": 102}, "e": {"f": [201, 301]}}}

    def test_file_modify(self) -> None:
        modder = Modder(actions=[FileActions])
        modder.modify({"_file_create": {"content": "Test data"}}, "test_file")
        modder.modify({"_file_copy": {"dest": "test_file_copy"}}, "test_file")
        modder.modify(
            {"_file_copy": {"dest1": "test_file_copy1", "dest2": "test_file_copy2"}},
            "test_file",
        )
        modder.modify({"_file_move": {"dest": "renamed_test_file"}}, "test_file")
        modder.modify({"_file_delete": {"mode": "actual"}}, "renamed_test_file")
        modder.modify({"_file_modify": {"mode": 0o666}}, "test_file_copy")
        modder.modify({"_file_delete": {"mode": "actual"}}, "test_file_copy")
        modder.modify({"_file_delete": {"mode": "actual"}}, "test_file_copy1")
        modder.modify({"_file_delete": {"mode": "actual"}}, "test_file_copy2")

    def test_strict_mode(self) -> None:
        modder = Modder(actions=[FileActions])
        dct = {"Hello": "World"}
        mod = {"_set": {"Hello": "Universe", "Bye": "World"}}
        with pytest.raises(ValueError, match="_set is not a supported action"):
            modder.modify(mod, dct)

        # In non-strict mode, unknown actions are ignored.
        dct = {"Hello": "World"}
        modder = Modder(actions=[FileActions], strict=False)
        modder.modify(mod, dct)
        assert dct == {"Hello": "World"}

        # File actions not supported
        modder = Modder()
        with pytest.raises(ValueError, match="_file_create is not a supported action"):
            modder.modify(
                {"_file_create": {"content": "Test data"}},
                "test_file",
            )

    def test_modify_object(self) -> None:
        modder = Modder()
        o = MyObject(1)
        assert o.b["a"] == 1
        mod_o = modder.modify_object({"_set": {"b->a": 20}}, o)
        assert mod_o.b["a"] == 20


class MyObject:
    def __init__(self, a) -> None:
        self.b = {"a": a}

    def as_dict(self):
        return {"b": {"a": self.b["a"]}}

    @staticmethod
    def from_dict(dct):
        return MyObject(dct["b"]["a"])



"""This package implements Lobster Jobs and Error Handlers."""

__author__ = "Janine George, Guido Petretto"
__version__ = "0.1"
__maintainer__ = "Janine George, Guido Petretto"
__email__ = "janine.george@uclouvain.be"
__status__ = "Production"
__date__ = "27/4/20"



"""This package implements various Gaussian Jobs and Error Handlers."""

__author__ = "Rasha Atwi"
__version__ = "0.1"
__maintainer__ = "Rasha Atwi"
__email__ = "rasha.atwi@stonybrook.edu"
__status__ = "Alpha"
__date__ = "5/13/21"



"""
This module implements a Modder class that performs modifications on objects
using support actions.
"""

import re

from custodian.ansible.actions import DictActions


class Modder:
    """
    Class to modify a dict/file/any object using a mongo-like language.
    Keywords are mostly adopted from mongo's syntax, but instead of $, an
    underscore precedes action keywords. This is so that the modification can
    be inserted into a mongo db easily.

    Allowable actions are supplied as a list of classes as an argument. Refer
    to the action classes on what the actions do. Action classes are in
    pymatpro.ansible.actions.

    Examples:
    >>> modder = Modder()
    >>> dct = {"Hello": "World"}
    >>> mod = {'_set': {'Hello':'Universe', 'Bye': 'World'}}
    >>> modder.modify(mod, dct)
    >>> dct['Bye']
    'World'
    >>> dct['Hello']
    'Universe'
    """

    def __init__(self, actions=None, strict=True, directory="./") -> None:
        """Initialize a Modder from a list of supported actions.

        Args:
            actions ([Action]): A sequence of supported actions. See
                :mod:`custodian.ansible.actions`. Default is None,
                which means only DictActions are supported.
            strict (bool): Indicating whether to use strict mode. In non-strict
                mode, unsupported actions are simply ignored without any
                errors raised. In strict mode, if an unsupported action is
                supplied, a ValueError is raised. Defaults to True.
            directory (str): The directory containing the files to be modified.
                Defaults to "./".
        """
        self.supported_actions = {}
        actions = actions if actions is not None else [DictActions]
        for action in actions:
            for attr in dir(action):
                if (not re.match(r"__\w+__", attr)) and callable(getattr(action, attr)):
                    self.supported_actions[f"_{attr}"] = getattr(action, attr)
        self.strict = strict
        self.directory = directory

    def modify(self, modification, obj) -> None:
        """
        Note that modify makes actual in-place modifications. It does not
        return a copy.

        Args:
            modification (dict): Modification must be {action_keyword :
                settings}. E.g., {'_set': {'Hello':'Universe', 'Bye': 'World'}}
            obj (dict/str/object): Object to modify depending on actions. For
                example, for DictActions, obj will be a dict to be modified.
                For FileActions, obj will be a string with a full pathname to a
                file.
        """
        for action, settings in modification.items():
            if action in self.supported_actions:
                self.supported_actions[action](obj, settings, directory=self.directory)
            elif self.strict:
                raise ValueError(f"{action} is not a supported action!")

    def modify_object(self, modification, obj):
        """
        Modify an object that supports pymatgen's as_dict() and from_dict API.

        Args:
            modification (dict): Modification must be {action_keyword :
                settings}. E.g., {'_set': {'Hello':'Universe', 'Bye': 'World'}}
            obj (object): Object to modify
        """
        dct = obj.as_dict()
        self.modify(modification, dct)
        return obj.from_dict(dct)


if __name__ == "__main__":
    import doctest

    doctest.testmod()
